#!/usr/bin/perl

###############
# OPEN TAGS   #
###############

my $zero_vector = pack("b*", "0"x32);

my %OPEN_TAGS = (
		 'B'  => $zero_vector,
		 'EM'  => $zero_vector,
		 'I'  => $zero_vector,
		 'PL' => $zero_vector,
		 'S'  => $zero_vector,
		 'TT' => $zero_vector,
		 'U'  => $zero_vector,
		 '0'  => $zero_vector,
		 '1'  => $zero_vector,
		 '2'  => $zero_vector,
		 '3'  => $zero_vector,
		 '4'  => $zero_vector,
		 '5'  => $zero_vector,
		 '6'  => $zero_vector,
		 '7'  => $zero_vector,
		 '8'  => $zero_vector,
		 '9'  => $zero_vector,
		 'r'  => $zero_vector,
		 'g'  => $zero_vector,
		 'b'  => $zero_vector,
		 'c'  => $zero_vector,
		 'm'  => $zero_vector,
		 'y'  => $zero_vector,
		 'k'  => $zero_vector,
		 'w'  => $zero_vector,
		);

vec($OPEN_TAGS{B}, 0, 1) = 1;
vec($OPEN_TAGS{EM}, 1, 1) = 1;
vec($OPEN_TAGS{I}, 2, 1) = 1;
vec($OPEN_TAGS{PL}, 3, 1) = 1;
vec($OPEN_TAGS{S}, 4, 1) = 1;
vec($OPEN_TAGS{TT}, 5, 1) = 1;
vec($OPEN_TAGS{U}, 6, 1) = 1;
vec($OPEN_TAGS{0}, 7, 1) = 1;
vec($OPEN_TAGS{1}, 8, 1) = 1;
vec($OPEN_TAGS{2}, 9, 1) = 1;
vec($OPEN_TAGS{3}, 10, 1) = 1;
vec($OPEN_TAGS{4}, 11, 1) = 1;
vec($OPEN_TAGS{5}, 12, 1) = 1;
vec($OPEN_TAGS{6}, 13, 1) = 1;
vec($OPEN_TAGS{7}, 14, 1) = 1;
vec($OPEN_TAGS{8}, 15, 1) = 1;
vec($OPEN_TAGS{9}, 16, 1) = 1;
vec($OPEN_TAGS{r}, 17, 1) = 1;
vec($OPEN_TAGS{g}, 18, 1) = 1;
vec($OPEN_TAGS{b}, 19, 1) = 1;
vec($OPEN_TAGS{c}, 20, 1) = 1;
vec($OPEN_TAGS{m}, 21, 1) = 1;
vec($OPEN_TAGS{y}, 22, 1) = 1;
vec($OPEN_TAGS{k}, 23, 1) = 1;
vec($OPEN_TAGS{w}, 24, 1) = 1;


###############
# CLOSE TAGS  #
###############

my $one_vector = pack("b*", "1"x32);

my %CLOSE_TAGS = (
		 'B'  => $one_vector,
		 'EM'  => $one_vector,
		 'I'  => $one_vector,
		 'PL' => $one_vector,
		 'S'  => $one_vector,
		 'TT' => $one_vector,
		 'U'  => $one_vector,
		 '0'  => $one_vector,
		 '1'  => $one_vector,
		 '2'  => $one_vector,
		 '3'  => $one_vector,
		 '4'  => $one_vector,
		 '5'  => $one_vector,
		 '6'  => $one_vector,
		 '7'  => $one_vector,
		 '8'  => $one_vector,
		 '9'  => $one_vector,
		 'r'  => $one_vector,
		 'g'  => $one_vector,
		 'b'  => $one_vector,
		 'c'  => $one_vector,
		 'm'  => $one_vector,
		 'y'  => $one_vector,
		 'k'  => $one_vector,
		 'w'  => $one_vector,
		);

vec($CLOSE_TAGS{B}, 0, 1) = 0;
vec($CLOSE_TAGS{EM}, 1, 1) = 0;
vec($CLOSE_TAGS{I}, 2, 1) = 0;
vec($CLOSE_TAGS{PL}, 3, 1) = 0;
vec($CLOSE_TAGS{S}, 4, 1) = 0;
vec($CLOSE_TAGS{TT}, 5, 1) = 0;
vec($CLOSE_TAGS{U}, 6, 1) = 0;
vec($CLOSE_TAGS{0}, 7, 1) = 0;
vec($CLOSE_TAGS{1}, 8, 1) = 0;
vec($CLOSE_TAGS{2}, 9, 1) = 0;
vec($CLOSE_TAGS{3}, 10, 1) = 0;
vec($CLOSE_TAGS{4}, 11, 1) = 0;
vec($CLOSE_TAGS{5}, 12, 1) = 0;
vec($CLOSE_TAGS{6}, 13, 1) = 0;
vec($CLOSE_TAGS{7}, 14, 1) = 0;
vec($CLOSE_TAGS{8}, 15, 1) = 0;
vec($CLOSE_TAGS{9}, 16, 1) = 0;
vec($CLOSE_TAGS{r}, 17, 1) = 0;
vec($CLOSE_TAGS{g}, 18, 1) = 0;
vec($CLOSE_TAGS{b}, 19, 1) = 0;
vec($CLOSE_TAGS{c}, 20, 1) = 0;
vec($CLOSE_TAGS{m}, 21, 1) = 0;
vec($CLOSE_TAGS{y}, 22, 1) = 0;
vec($CLOSE_TAGS{k}, 23, 1) = 0;
vec($CLOSE_TAGS{w}, 24, 1) = 0;


my %OPEN_CLOSE;
foreach (keys %OPEN_TAGS) {
  $OPEN_CLOSE{$OPEN_TAGS{$_}} = $CLOSE_TAGS{$_};
}

my %INV_TAGS;
foreach (keys %OPEN_TAGS) {
  $INV_TAGS{$OPEN_TAGS{$_}} = $_;
}

foreach (keys %CLOSE_TAGS) {
  $INV_TAGS{$CLOSE_TAGS{$_}} = "/".$_;
}

####################################################
$|= 1;

my $context = $zero_vector;

my $c;
my @buffer;

while (read(STDIN, $c, 1)) {
  if ($c eq "<") {
    parse_tag();
  } else {
    $c =~ s/\t|\s|\n|\r/ /g;
    push @buffer, $c;
  }
}

#print "---------------\n";

my $string = join "\n", @buffer;
#print $string;
#print "\n--\n";
simplify_context();
@buffer = split "\n", $string;
#print $string;
#print "\n--\n";
rewrite();
#print "\n--\n";


my %spares;
sub parse_tag {
  my $c;
  my $tag;

  # is this an opening tag?
  my $open = 1;
  read(STDIN, $c, 1);
  if ($c eq "/") {
    $open = 0;
  } else {
    $tag = $c;
  }
  
  while (not ($c eq ">")) {
    read(STDIN, $c, 1);
    $tag .= $c;
  }
  
  chop($tag);
  
  if ($open) {
    if (tag_in_context($OPEN_TAGS{$tag}, $context)) {
      # that's a redundant tag
      $spares{$tag}++;
    } else {
      $context = $context | $OPEN_TAGS{$tag};
      my $vec = unpack("b*", $context);
      push @buffer, $vec;
    }
  } else {
    if ($spares{$tag}) {
      $spares{$tag}--;
    } else {
      $context = $context & $CLOSE_TAGS{$tag};
      my $vec = unpack("b*", $context);
      push @buffer, $vec;
    }
  }
  
}

sub simplify_context {
  $string =~ s/(\d{32}\n)+(\d{32}\n)/$2/g;
}


my @close_tags_stack;
sub rewrite {
  my $i = 0;

  my $last_context = $zero_vector;
  my $current_context = $zero_vector;
  my $last_was_space = 0;

  my $last_space_context = space_context($last_context);
  my $current_space_context = space_context($current_context);

  foreach (@buffer) {
    if (/^\d{32}$/) { # whenever we encounter a change in context
      $current_context = pack("b*", $buffer[$i]);
      $current_space_context = space_context($current_context);

      my @to_close = closing_tags($last_context, $current_context);
      while (@to_close) {
	my $cl_tag = pop @close_tags_stack;
	if (tag_in_set($cl_tag, \@to_close)) {
	  del_tag_from_set($cl_tag, \@to_close);
	} else {
	  # unset the tag in the context so that it opens again
	  $last_context = $last_context & $cl_tag;
	}
	print_tag($cl_tag);
      }
      
      
      # find opening tags
      my %itl;
      foreach (opening_tags($last_context, $current_context)) {
	 $itl{$_} = itl($i, $_);
      }

      # sort the opening tags by their duration
      foreach (sort {$itl{$b} <=> $itl{$a}} keys %itl) {
	# print the opening tag
	print_tag($_);

	# save the closing tag
	push @close_tags_stack, $OPEN_CLOSE{$_};
      }
      
      $last_context = $current_context;
    } else {
      # this is a printable char
#      print "\n";
#      print "\n";
#      print unpack("b*", $last_space_context);
#      print "\n";
#      print unpack("b*", $current_space_context);
#      print "\n";
      if (m/\n|\t|\s|\r/) { # check if this is a space character
	if (unpack("b*", $last_space_context) eq unpack("b*", $current_space_context) and $last_was_space) { # only print if the space context has changed
	  #print "<NOSPC>";
	} else {
	  print;
	  # space context
	  $last_space_context = $current_space_context;
	}
	$last_was_space = 1;
      } else {
	print;
	$last_was_space = 0;
      }
    }
    
    $i++;
  }
}

sub space_context {
  my $current_context = shift;

  # unset S, EM, I, B
  $current_context = $current_context & $CLOSE_TAGS{S};
  $current_context = $current_context & $CLOSE_TAGS{EM};
  $current_context = $current_context & $CLOSE_TAGS{I};
  $current_context = $current_context & $CLOSE_TAGS{B};

  # if U is not set then set the color to w
  if (not tag_in_context($OPEN_TAGS{U}, $current_context)) {
    $current_context = $current_context & $CLOSE_TAGS{r};
    $current_context = $current_context & $CLOSE_TAGS{g};
    $current_context = $current_context & $CLOSE_TAGS{b};
    $current_context = $current_context & $CLOSE_TAGS{c};
    $current_context = $current_context & $CLOSE_TAGS{m};
    $current_context = $current_context & $CLOSE_TAGS{y};
    $current_context = $current_context & $CLOSE_TAGS{k};
    $current_context = $current_context | $OPEN_TAGS{w};
  }
      
  # if no color is set the set the color to default w for space context
  unless (
	  tag_in_context($OPEN_TAGS{r}, $current_context)
	  or tag_in_context($OPEN_TAGS{g}, $current_context)
	  or tag_in_context($OPEN_TAGS{b}, $current_context)
	  or tag_in_context($OPEN_TAGS{c}, $current_context)
	  or tag_in_context($OPEN_TAGS{m}, $current_context)
	  or tag_in_context($OPEN_TAGS{y}, $current_context)
	  or tag_in_context($OPEN_TAGS{k}, $current_context)
	 ) {
    $current_context = $current_context | $OPEN_TAGS{w};
  }
      

  return $current_context;
}

sub del_tag_from_set {
  my $tag = shift;
  my $set_ref = shift;

  my $i = 0;
  my $offset = -1;
  my $utag = unpack("b*", $tag);
  foreach (@{$set_ref}) {
    $offset = $i if unpack("b*", $_) eq $utag;
    $i++;
  }
  
  unless ($offset == -1) {
    splice @{$set_ref}, $offset, 1;
  }
}  

sub tag_in_set {
  my $tag = shift;
  my $set_ref = shift;

  my $utag = unpack("b*", $tag);
  foreach (@{$set_ref}) {
    return 1 if unpack("b*", $_) eq $utag;
  }
  return 0;
}

sub print_tag {
  my $tag = shift;
  print "<".$INV_TAGS{$tag}.">";
}

sub itl { 
  # iterations to live: the tags with greater itl are written first
  # since they will be closed last this is not always optimal though

  my $step = shift;
  my $tag  = shift;
  
  my $itl = 0;
  my $i;
  for ($i = $step + 1; $i <= $#buffer; $i++) {
    if ($buffer[$i] =~ m/\d{32}/) {
      my $context = pack("b*", $buffer[$i]);
      return $itl unless tag_in_context($tag, $context);
    } else {
      $itl++;
    }
  }
  
  return 0;
}

sub opening_tags {
  my $context_1 = shift;
  my $context_2 = shift;

  my @opening_tags;
  foreach (keys %OPEN_TAGS) {
    if (not tag_in_context($OPEN_TAGS{$_}, $context_1) and tag_in_context($OPEN_TAGS{$_}, $context_2)) {
      push @opening_tags, $OPEN_TAGS{$_};
    }
  }
  
  return @opening_tags;
}  

sub closing_tags {
  my $context_1 = shift;
  my $context_2 = shift;

  my @closing_tags;
  foreach (keys %OPEN_TAGS) {
    if (tag_in_context($OPEN_TAGS{$_}, $context_1) and not tag_in_context($OPEN_TAGS{$_}, $context_2)) {
      push @closing_tags, $OPEN_CLOSE{$OPEN_TAGS{$_}};
    }
  }
  
  return @closing_tags;
}

sub active_tags { # tags with bits set in curreny context
  my $context = shift;

  my @active;
  my $i;
  for ($i = 0; $i < 25; $i++) {
    my $tag = $zero_vector;
    vec($tag, $i, 1) = 1;
    if (tag_in_context($tag, $context)) {
      push @active, $tag;
    }
  }
  
  return @active;
}

sub tag_in_context { # is tag active in current context?
  my $tag = shift;
  my $context = shift;

  return (unpack("b*", $tag & $context) eq unpack("b*", $tag))  
}
